package be.iminds.ilabt.jfed.rspec;

import be.iminds.ilabt.jfed.rspec.model.*;
import be.iminds.ilabt.jfed.rspec.rspec_source.AdvertisementRspecSource;
import be.iminds.ilabt.jfed.rspec.rspec_source.RequestRspecSource;
import com.fasterxml.jackson.annotation.JsonIgnoreProperties;
import com.fasterxml.jackson.annotation.JsonInclude;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.*;
import java.util.stream.Collectors;

/**
 * Generate an Rspec based on info in an advertisement RSpec
 */
@JsonInclude(JsonInclude.Include.NON_DEFAULT)
@JsonIgnoreProperties(value={"@context"})
public class AdvertisementBasedRspecGenerator {
    private static final Logger LOG = LoggerFactory.getLogger(AdvertisementBasedRspecGenerator.class);

    @Nonnull private final RspecFactory rspecFactory;
    @Nonnull private final AdvertisementBasedRspecGeneratorConfig config;
    @Nonnull private final AdvertisementRspecSource advRspecSource;

    public AdvertisementBasedRspecGenerator(@Nonnull RspecFactory rspecFactory,
                                            @Nonnull AdvertisementBasedRspecGeneratorConfig config,
                                            @Nonnull AdvertisementRspecSource advRspecSource) {
        this.rspecFactory = rspecFactory;
        this.config = config;
        this.advRspecSource = advRspecSource;
    }

    public RequestRspecSource generate() {
        assert advRspecSource != null;
        ModelRspec model = advRspecSource.getModelRspec();
        if (model == null)
            throw new RspecGeneratorException("Problem with advertisement RSpec");

        LinkedList<RspecNode> unusedNodes = new LinkedList<>(model.getNodes());

        ModelRspec res = rspecFactory.createModelRspec("request");

        for (AdvertisementBasedRspecGeneratorConfig.NodeInfo nodeInfo : config.getNodes()) {
            for (int i = 0; i < nodeInfo.getCount(); i++) {
                String compIdSelector = (nodeInfo.getComponentIdSelector() == null
                        || nodeInfo.getComponentIdSelector().isEmpty() ?
                        null :
                        (i < nodeInfo.getComponentIdSelector().size() ?
                                nodeInfo.getComponentIdSelector().get(i) :
                                nodeInfo.getComponentIdSelector().get(nodeInfo.getComponentIdSelector().size()-1)));
                RspecNode n = findNode(unusedNodes,
                        nodeInfo.getAvailableFilter(),
                        nodeInfo.getSliverType(),
                        nodeInfo.getHardwareTypeFilter(),
                        compIdSelector,
                        nodeInfo.getComponentIdFilter(),
                        nodeInfo.getExclusive());
                if (n == null) {
                    throw new RspecGeneratorException("Failed to find matching node");
                }

                RspecNode newNode = rspecFactory.createNode(res);
                newNode.setComponentManagerId(n.getComponentManagerId());
                newNode.setExclusive(n.getExclusive());
                newNode.setComponentId(n.getComponentId());
                newNode.setSliverTypeName(nodeInfo.sliverType == null ?
                        n.getSliverTypes().iterator().next().getName() :
                        nodeInfo.sliverType);
                newNode.setClientId(createUniqueName(nodeInfo, res));
                if (nodeInfo.getDiskImage() != null)
                    newNode.setSliverTypeDiskImage(nodeInfo.getDiskImage());
                if (nodeInfo.getAnsibleGroup() != null)
                    newNode.getAnsibleGroups().add(nodeInfo.getAnsibleGroup());
                res.addNode(newNode);
            }
        }

        return new RequestRspecSource(res);
    }

    private String createUniqueName(@Nonnull AdvertisementBasedRspecGeneratorConfig.NodeInfo nodeInfo,
                                    @Nonnull ModelRspec res) {
        boolean nameUsedForMultipleNodes = nodeInfo.getCount() > 1;
        String clientIdBase = nodeInfo.getClientIdBase();

        if (!nameUsedForMultipleNodes) {
            //see if any other node info uses the same base name
            for (AdvertisementBasedRspecGeneratorConfig.NodeInfo curNodeInfo : config.getNodes()) {
                //noinspection ObjectEquality
                if (curNodeInfo == nodeInfo)
                    continue;
                if (curNodeInfo.getClientIdBase().equals(clientIdBase))
                    nameUsedForMultipleNodes = true;
            }
        }

        if (res.getNodes().size() == 0) {
            return nameUsedForMultipleNodes ?
                    clientIdBase+""+1 :
                    clientIdBase;
        }

        for (int i = 1; i <= res.getNodes().size()+1; i++) {
            String proposedName = i == 1 && !nameUsedForMultipleNodes ?
                    clientIdBase :
                    clientIdBase+""+i;

            boolean proposedNameTaken = false;
            for (RspecNode existingNode : res.getNodes()) {
                if (proposedName.equals(existingNode.getClientId()))
                    proposedNameTaken = true;
            }
            if (!proposedNameTaken)
                return proposedName;
        }

        LOG.warn("Bug in createUniqueName. node names = ["+
                res.getNodes().stream()
                        .map(RspecNode::getClientId)
                        .collect(Collectors.joining(", "))+
                "] clientIdBase="+clientIdBase+
                " nameUsedForMultipleNodes="+nameUsedForMultipleNodes);

        //should never happen, but safe fallback anyway
        return clientIdBase+"-error-"+Math.random();
    }

    private Random random = new Random();
    private RspecNode findNode(@Nonnull LinkedList<RspecNode> nodes,
                               @Nullable Boolean available,
                               @Nullable String sliverType,
                               @Nullable List<String> hardwareTypeFilter,
                               @Nullable String componentIdSelector /*urn, or "random" or "alphabetically"*/,
                               @Nullable List<String> componentIdFilter,
                               Boolean exclusive) {
        LOG.debug("Looking for node in list of "+nodes.size());
        if (nodes.isEmpty())
            return null;

        SortedSet<RspecNode> matchingNodes = new TreeSet<>(new Comparator<RspecNode>() {
            @Override
            public int compare(RspecNode a, RspecNode b) {
                return a.getUniqueId().compareTo(b.getUniqueId());
            }
        });

        for (ListIterator<RspecNode> it = nodes.listIterator(); it.hasNext(); ) {
            RspecNode curNode = it.next();
            if (available != null
                    && curNode.getAvailable() != null
                    && !curNode.getAvailable().equals(available)) {
                LOG.debug("Node skipped due to available");
                continue;
            }
            if (componentIdSelector != null &&
                    componentIdSelector.startsWith("urn:") && (
                    curNode.getComponentId() == null ||
                            !curNode.getComponentId().toString().equals(componentIdSelector))
                    ) {
                LOG.debug("Node skipped due to componentIdSelector");
                continue;
            }
            if (hardwareTypeFilter != null &&
                    !hardwareTypeFilter.isEmpty() &&
                    (
                            curNode.getHardwareTypes() == null ||
                                    curNode.getHardwareTypes().stream()
                                            .map(HardwareType::getName)
                                            .filter(Objects::nonNull)
                                            .noneMatch(hardwareTypeFilter::contains)
                    )
                    ) {
                LOG.debug("Node skipped due to hardwareTypeFilter");
                continue;
            }
            if (componentIdFilter != null &&
                    !componentIdFilter.isEmpty() &&
                    !componentIdFilter.contains(curNode.getComponentId().toString())
                    ) {
                LOG.debug("Node skipped due to componentIdFilter");
                continue;
            }
            //slivertype is also used a filter if present
            if (sliverType != null &&
                    !sliverType.isEmpty() &&
                    (
                            curNode.getHardwareTypes() == null ||
                                    curNode.getSliverTypes().stream()
                                            .map(SliverType::getName)
                                            .filter(Objects::nonNull)
                                            .noneMatch(sliverType::equals)
                    )) {
                LOG.debug("Node skipped due to sliverType");
                continue;
            }
            //exclusive is also used a filter if present
            if (exclusive != null &&
                    exclusive &&
                    (curNode.getExclusive() == null
                            || Objects.equals(curNode.getExclusive(), Boolean.FALSE))) {
                LOG.debug("Node skipped due to exclusive");
                continue;
            }

            it.remove();
            matchingNodes.add(curNode);
        }

        if (matchingNodes.isEmpty()) {
            LOG.debug("No matching nodes found");
            return null;
        }

        int chosenIndex = 0;
        if (componentIdSelector != null && componentIdSelector.equalsIgnoreCase("random")) {
            chosenIndex = random.nextInt(matchingNodes.size());
        }

        LinkedList<RspecNode> matchingNodesList = new LinkedList<>(matchingNodes);
        RspecNode res = matchingNodesList.remove(chosenIndex);
        nodes.addAll(matchingNodesList);
        if ((exclusive == null &&
                Objects.equals(res.getExclusive(), Boolean.FALSE)) ||
                Objects.equals(exclusive, Boolean.FALSE) ) {
            //don't remove node if we don't get exclusive access to it.
            //this means we'll reuse it for other request nodes as well.
            nodes.add(res);
            LOG.debug("Node is not exclusive -> can use again");
        } else {
            LOG.debug("Node is exclusive ("+exclusive+" "+res.getExclusive()+") -> will not use again");
        }
        return res;
    }

    public static class RspecGeneratorException extends RuntimeException {
        public RspecGeneratorException() { }
        public RspecGeneratorException(String s) { super(s); }
        public RspecGeneratorException(String s, Throwable throwable) { super(s, throwable); }
        public RspecGeneratorException(Throwable throwable) { super(throwable); }
        public RspecGeneratorException(String s, Throwable throwable, boolean b, boolean b1) { super(s, throwable, b, b1); }
    }
}
